"""Infer the constraints on the length of a property value."""
from typing import MutableMapping, Optional, Tuple, List, Union, Sequence

from icontract import require, ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    assert_never,
    Error,
    Identifier,
    assert_union_of_descendants_exhaustive,
)
from aas_core_codegen.infer_for_schema import _common as infer_for_schema_common
from aas_core_codegen.parse import tree as parse_tree


class _Constraint:
    """
    Represent a constraint on the ``len`` of something.

    That something can be, *e.g.*, a property or ``self``.
    """

    def __init__(self, node: parse_tree.Node) -> None:
        self.node = node


class _MinLength(_Constraint):
    """Represent the constraint that the ``len`` is ≥ ``value``."""

    def __init__(self, node: parse_tree.Node, value: int) -> None:
        """Initialize with the given values."""
        _Constraint.__init__(self, node=node)
        self.value = value


class _MaxLength(_Constraint):
    """Represent the constraint that the ``len`` is ≤ ``value``."""

    def __init__(self, node: parse_tree.Node, value: int) -> None:
        """Initialize with the given values."""
        _Constraint.__init__(self, node=node)
        self.value = value


class _ExactLength(_Constraint):
    """Represent the constraint that the ``len`` is == ``value``."""

    def __init__(self, node: parse_tree.Node, value: int) -> None:
        """Initialize with the given values."""
        _Constraint.__init__(self, node=node)
        self.value = value


class _LenOnMemberOrName:
    """Represent a match on a call to ``len(.)`` on a member or a name."""

    def __init__(
        self, member_or_name: Union[parse_tree.Member, parse_tree.Name]
    ) -> None:
        """Initialize with the given values."""
        self.member_or_name = member_or_name


def _match_len_on_member_or_name(node: parse_tree.Node) -> Optional[_LenOnMemberOrName]:
    """
    Match expressions like ``len(self.something)``.

    Return the name of the property, or None, if not matched.
    """
    mtch = infer_for_schema_common.match_single_arg_function_on_member_or_name(node)

    if mtch is None:
        return None

    if mtch.function_name == "len":
        return _LenOnMemberOrName(member_or_name=mtch.member_or_name)

    return None


def _match_int_constant(node: parse_tree.Node) -> Optional[int]:
    """Match an integer constant."""
    if isinstance(node, parse_tree.Constant) and isinstance(node.value, int):
        return node.value

    return None


class _LenConstraintOnMemberOrName:
    """
    Represent a match on an expression like ``len(.) < 42``.

    Unlike :py:class:`LenConstraint`, which represents an *inferred* constrained over
    multiple constraints, this class represents a constraint as it is derived directly
    from the parse tree.
    """

    def __init__(
        self,
        member_or_name: Union[parse_tree.Member, parse_tree.Name],
        constraint: "_ConstraintUnion",
    ) -> None:
        """Initialize with the given values."""
        self.member_or_name = member_or_name
        self.constraint = constraint


def _match_len_constraint_on_member_or_name(
    node: parse_tree.Node,
) -> Optional[_LenConstraintOnMemberOrName]:
    """
    Match the constraint on ``len`` of a member or a variable.

    For example, something like ``len(self.something) < 42`` or ``len(self) == 42``.

    Return the match, or None if not matched.
    """
    if not isinstance(node, parse_tree.Comparison):
        return None

    constraint = None  # type: Optional[_ConstraintUnion]

    # region Match a comparison like ``len(self.something) < 42``

    len_on_member_or_name = _match_len_on_member_or_name(node.left)
    constant = _match_int_constant(node.right)

    if len_on_member_or_name is not None and constant is not None:
        if node.op is parse_tree.Comparator.LT:
            # len(self.something) < 42
            constraint = _MaxLength(node=node, value=constant - 1)
        elif node.op is parse_tree.Comparator.LE:
            # len(self.something) <= 42
            constraint = _MaxLength(node=node, value=constant)
        elif node.op is parse_tree.Comparator.EQ:
            # len(self.something) == 42
            constraint = _ExactLength(node=node, value=constant)
        elif node.op is parse_tree.Comparator.GT:
            # len(self.something) > 42
            constraint = _MinLength(node=node, value=constant + 1)
        elif node.op is parse_tree.Comparator.GE:
            # len(self.something) >= 42
            constraint = _MinLength(node=node, value=constant)
        elif node.op is parse_tree.Comparator.NE:
            # We intentionally ignore the invariants such as ``len(n) != 42``
            # as there is no meaningful way to represent it simply in a schema.
            pass
        else:
            assert_never(node.op)

        if constraint is not None:
            return _LenConstraintOnMemberOrName(
                member_or_name=len_on_member_or_name.member_or_name,
                constraint=constraint,
            )

    # endregion

    # region Match a comparison like ``42 < self.something``

    constant = _match_int_constant(node.left)
    len_on_member_or_name = _match_len_on_member_or_name(node.right)

    if constant is not None and len_on_member_or_name is not None:
        if node.op is parse_tree.Comparator.LT:
            # 42 < len(self.something)
            constraint = _MinLength(node=node, value=constant + 1)
        elif node.op is parse_tree.Comparator.LE:
            # 42 <= len(self.something)
            constraint = _MinLength(node=node, value=constant)
        elif node.op is parse_tree.Comparator.EQ:
            # 42 == len(self.something)
            constraint = _ExactLength(node=node, value=constant)
        elif node.op is parse_tree.Comparator.GT:
            # 42 > len(self.something)
            constraint = _MaxLength(node=node, value=constant - 1)
        elif node.op is parse_tree.Comparator.GE:
            # 42 >= len(self.something)
            constraint = _MaxLength(node=node, value=constant)
        elif node.op is parse_tree.Comparator.NE:
            # We intentionally ignore the invariants such as ``len(n) != 42``
            # as there is no meaningful way to represent it simply in a schema.
            pass
        else:
            assert_never(node.op)

        if constraint is not None:
            return _LenConstraintOnMemberOrName(
                member_or_name=len_on_member_or_name.member_or_name,
                constraint=constraint,
            )

    # endregion

    return None


def min_with_none(*args: Optional[int]) -> Optional[int]:
    """
    Compute a minimum among the arguments where None are ignored.

    >>> min_with_none(2, None, 1)
    1

    >>> min_with_none(None, None)
    """
    # NOTE (mristin, 2022-03-2):
    # There is no one-liner in Python for this.
    # See: https://stackoverflow.com/questions/2295461/list-minimum-in-python-with-none
    minimum = None  # type: Optional[int]
    for arg in args:
        if minimum is None:
            minimum = arg
        else:
            if arg is not None:
                minimum = min(arg, minimum)

    return minimum


def max_with_none(*args: Optional[int]) -> Optional[int]:
    """
    Compute a maximum among the arguments where None are ignored.

    >>> max_with_none(2, None, 1)
    2

    >>> max_with_none(None, None)
    """
    # NOTE (mristin, 2022-03-2):
    # There is no one-liner in Python for this.
    # See: https://stackoverflow.com/questions/2295461/list-minimum-in-python-with-none
    maximum = None  # type: Optional[int]
    for arg in args:
        if maximum is None:
            maximum = arg
        else:
            if arg is not None:
                maximum = max(arg, maximum)

    return maximum


class LenConstraint:
    """
    Represent the inferred constraint on the ``len`` of something.

    Both bounds are inclusive: ``min_value ≤ len ≤ max_value``.
    """

    # fmt: off
    @require(
        lambda min_value, max_value:
        not (min_value is not None and max_value is not None)
        or 0 < min_value <= max_value
    )
    # fmt: on
    def __init__(self, min_value: Optional[int], max_value: Optional[int]) -> None:
        """Initialize with the given values."""
        self.min_value = min_value
        self.max_value = max_value

    def copy(self) -> "LenConstraint":
        """Create a copy of the self."""
        return LenConstraint(min_value=self.min_value, max_value=self.max_value)

    def __str__(self) -> str:
        return (
            f"LenConstraint(min_value={self.min_value!r}, max_value={self.max_value!r})"
        )


class _LenConstraintOnProperty:
    """Represent a len constraint such as ``len(self.something) < 42``."""

    def __init__(self, prop_name: Identifier, constraint: "_ConstraintUnion") -> None:
        """Initialize with the given values."""
        self.prop_name = prop_name
        self.constraint = constraint


def _match_len_constraint_on_property(
    node: parse_tree.Expression,
) -> Optional[_LenConstraintOnProperty]:
    """Match a len constraint on a property such as ``len(self.something) < 42``."""
    len_constraint_on_member_or_name = _match_len_constraint_on_member_or_name(node)

    if len_constraint_on_member_or_name:
        prop_name = infer_for_schema_common.match_property(
            len_constraint_on_member_or_name.member_or_name
        )

        if prop_name is not None:
            return _LenConstraintOnProperty(
                prop_name=prop_name,
                constraint=len_constraint_on_member_or_name.constraint,
            )

    return None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _reduce_constraints(
    constraints: Sequence["_ConstraintUnion"],
) -> Tuple[Optional[LenConstraint], Optional[List[str]]]:
    """
    Reduce a list of constraints to a range that encompasses all the constraints.

    Return the reduced constraint, or a list of errors.
    """
    min_len = None  # type: Optional[int]
    max_len = None  # type: Optional[int]
    exact_len = None  # type: Optional[int]

    errors = []  # type: List[str]

    for constraint in constraints:
        if isinstance(constraint, _MinLength):
            min_len = max_with_none(constraint.value, min_len)

        elif isinstance(constraint, _MaxLength):
            max_len = min_with_none(constraint.value, max_len)

        elif isinstance(constraint, _ExactLength):
            if exact_len is not None:
                errors.append(
                    f"The exact length, {exact_len}, contradicts "
                    f"another exactly expected length {constraint.value}.",
                )
            exact_len = constraint.value

        else:
            assert_never(constraint)

    if exact_len is not None:
        if min_len is not None and min_len > exact_len:
            errors.append(
                f"the minimum length, {min_len}, "
                f"contradicts the exactly expected length {exact_len}.",
            )

        if max_len is not None and exact_len > max_len:
            errors.append(
                f"the maximum length, {max_len}, "
                f"contradicts the exactly expected length {exact_len}.",
            )

    if min_len is not None and max_len is not None and min_len > max_len:
        errors.append(
            f"the minimum length, {min_len}, "
            f"contradicts the maximum length {max_len}.",
        )

    if len(errors) > 0:
        return None, errors

    if exact_len is not None:
        min_len = exact_len
        max_len = exact_len

    return LenConstraint(min_value=min_len, max_value=max_len), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def len_constraints_from_invariants(
    cls: intermediate.Class,
) -> Tuple[
    Optional[MutableMapping[intermediate.Property, LenConstraint]],
    Optional[List[Error]],
]:
    """
    Infer the constraints on ``len`` for every property of the class ``cls``.

    Even if a property is optional, the constraint will still be inferred. Please be
    careful that this does not scramble your cardinality constraints (which start from
    zero for optional properties).

    The constraints are not exhaustive. We only infer constraints based on invariants
    which involve constants. It might be that the actual invariants are tighter.
    """
    constraint_map = dict()  # type: MutableMapping[Identifier, List[_ConstraintUnion]]

    # region Infer the constraints in the loose form from all the invariants

    # NOTE (mristin, 2021-11-30):
    # We iterate only once through the invariants instead of inferring the constraints
    # for each property individually to be able to keep linear time complexity.

    errors = []  # type: List[Error]

    for invariant in cls.invariants:
        # NOTE (mristin, 2022-01-02):
        # We consider only the genuine invariants of the class, and ignore
        # the invariants of its ancestors.

        if invariant.specified_for is not cls:
            continue

        # noinspection PyUnusedLocal
        len_constraint_on_prop = None  # type: Optional[_LenConstraintOnProperty]

        # Match ``self.something is None or len(self.something) < X``
        conditional_on_prop = infer_for_schema_common.match_conditional_on_prop(
            invariant.body
        )
        if conditional_on_prop is not None:
            len_constraint_on_prop = _match_len_constraint_on_property(
                conditional_on_prop.consequent
            )

        else:
            # Match ``len(self.something) < X``
            len_constraint_on_prop = _match_len_constraint_on_property(invariant.body)

        if len_constraint_on_prop is not None:
            if len_constraint_on_prop.prop_name not in cls.properties_by_name:
                errors.append(
                    Error(
                        invariant.body.original_node,
                        f"The property {len_constraint_on_prop.prop_name} does not "
                        f"appear in the properties of the class {cls.name}",
                    )
                )
                continue

            constraints = constraint_map.get(len_constraint_on_prop.prop_name, None)
            if constraints is None:
                constraints = []
                constraint_map[len_constraint_on_prop.prop_name] = constraints

            constraints.append(len_constraint_on_prop.constraint)

    # endregion

    if len(errors) > 0:
        return None, errors

    # region Compress the loose constraints

    result = dict()  # type: MutableMapping[intermediate.Property, LenConstraint]

    for prop_name, constraints in constraint_map.items():
        reduced_constraint, reduction_errors = _reduce_constraints(
            constraints=constraints
        )

        if reduction_errors is not None:
            for reduction_error in reduction_errors:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The property {prop_name} has conflicting invariants "
                        f"on the length: {reduction_error}",
                    )
                )
            continue

        assert reduced_constraint is not None

        prop = cls.properties_by_name.get(prop_name, None)
        assert prop is not None, (
            f"Expected the property {prop_name!r} in the properties "
            f"of the symbol {cls}"
        )

        result[prop] = reduced_constraint

    # endregion

    if len(errors) > 0:
        return None, errors

    return result, None


LENGTHABLE_PRIMITIVES = frozenset(
    [intermediate.PrimitiveType.STR, intermediate.PrimitiveType.BYTEARRAY]
)


# fmt: off
@require(
    lambda constrained_primitive:
    constrained_primitive.constrainee in LENGTHABLE_PRIMITIVES,
    "The length inferred only for meaningful primitives"
)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def infer_len_constraint_of_self(
    constrained_primitive: intermediate.ConstrainedPrimitive,
) -> Tuple[Optional[LenConstraint], Optional[List[Error]],]:
    """
    Infer the constraint on ``len(self)``.

    The constraints are not exhaustive. We only infer constraints based on invariants
    which involve constants. It might be that the actual invariants are tighter.
    """
    # region Infer the constraints in the loose form from all the invariants

    errors = []  # type: List[Error]

    constraints = []  # type: List[_ConstraintUnion]

    for invariant in constrained_primitive.invariants:
        # NOTE (mristin, 2022-01-02):
        # We consider only the genuine invariants of the constrained primitive, and
        # ignore the invariants of its ancestors.
        if invariant.specified_for is not constrained_primitive:
            continue

        # Match something like ``len(self) < 42``
        len_constraint_on_member_or_name = _match_len_constraint_on_member_or_name(
            invariant.body
        )

        if len_constraint_on_member_or_name:
            # Abbreviate for readability
            member_or_name = len_constraint_on_member_or_name.member_or_name

            if isinstance(member_or_name, parse_tree.Name) and (
                member_or_name.identifier == "self"
            ):
                constraints.append(len_constraint_on_member_or_name.constraint)

        # Otherwise, ignore the invariant as we do not understand it.

    # endregion

    if len(errors) > 0:
        return None, errors

    # region Compress the loose constraints

    reduced_constraint, reduction_errors = _reduce_constraints(constraints=constraints)

    if reduction_errors is not None:
        for reduction_error in reduction_errors:
            errors.append(
                Error(
                    constrained_primitive.parsed.node,
                    f"There are conflicting invariants on the length: "
                    f"{reduction_error}",
                )
            )
        return None, errors

    assert reduced_constraint is not None

    # endregion

    return reduced_constraint, None


_ConstraintUnion = Union[_MinLength, _MaxLength, _ExactLength]
assert_union_of_descendants_exhaustive(union=_ConstraintUnion, base_class=_Constraint)



# pylint: disable=missing-docstring
# pylint: disable=no-self-use

import textwrap
import unittest

import tests.common
from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference


class Test_with_smoke(unittest.TestCase):
    @staticmethod
    def execute(source: str) -> None:
        """Execute a smoke test on all the invariants of all the classes."""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        base_environment = intermediate_type_inference.populate_base_environment(
            symbol_table=symbol_table
        )

        for symbol in symbol_table.symbols:
            if isinstance(symbol, intermediate.Class):
                environment = intermediate_type_inference.MutableEnvironment(
                    parent=base_environment
                )
                environment.set(
                    Identifier("self"),
                    intermediate_type_inference.OurTypeAnnotation(
                        symbol=symbol_table.must_find(Identifier("Some_class"))
                    ),
                )

                for invariant in symbol.invariants:
                    inferrer = intermediate_type_inference.Inferrer(
                        symbol_table=symbol_table, environment=environment
                    )

                    inferrer.transform(invariant.body)
                    assert (
                        len(inferrer.errors) == 0
                    ), tests.common.most_underlying_messages(inferrer.errors)

    def test_enumeration_literal_as_member(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                Literal_a = "LITERAL-A"
                Literal_b = "LITERAL-B"
                Literal_c = "LITERAL-C"

            @invariant(
                lambda self:
                self.something == Some_enum.Literal_a
                or self.something == Some_enum.Literal_b
            )
            class Some_class:
                something: Some_enum

                def __init__(self, something: Some_enum) -> None:
                    self.something = something


            __book_url__ = "dummy"
            __book_version__ = "dummy"
            """
        )

        Test_with_smoke.execute(source=source)

    def test_class_member(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(
                lambda self:
                self.something > 0
            )
            class Some_class:
                something: int

                def __init__(self, something: int) -> None:
                    self.something = something


            __book_url__ = "dummy"
            __book_version__ = "dummy"
            """
        )

        Test_with_smoke.execute(source=source)


if __name__ == "__main__":
    unittest.main()



"""Stringify the intermediate representation."""

from typing import Union, Optional

from aas_core_codegen import stringify
from aas_core_codegen.intermediate import _types, construction
from aas_core_codegen.intermediate._types import (
    AbstractClass,
    Argument,
    Class,
    ConcreteClass,
    ConstrainedPrimitive,
    Constructor,
    Contract,
    Contracts,
    Default,
    DefaultConstant,
    DefaultEnumerationLiteral,
    Enumeration,
    EnumerationLiteral,
    ImplementationSpecificMethod,
    ImplementationSpecificVerification,
    Interface,
    Invariant,
    ListTypeAnnotation,
    MetaModel,
    Method,
    OptionalTypeAnnotation,
    OurTypeAnnotation,
    PatternVerification,
    PrimitiveTypeAnnotation,
    Property,
    Serialization,
    Signature,
    Snapshot,
    Symbol,
    SymbolTable,
    UnderstoodMethod,
    ReferenceInTheBook,
    MetaModelDescription,
    SymbolDescription,
    SignatureDescription,
    PropertyDescription,
    EnumerationLiteralDescription,
    SignatureLike,
)
from aas_core_codegen.parse import tree as parse_tree


def _stringify_primitive_type_annotation(
    that: PrimitiveTypeAnnotation,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("a_type", that.a_type.name),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_our_type_annotation(
    that: OurTypeAnnotation,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("symbol", f"Reference to symbol {that.symbol.name}"),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_list_type_annotation(
    that: ListTypeAnnotation,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("items", _stringify(that.items)),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_optional_type_annotation(
    that: OptionalTypeAnnotation,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("value", _stringify(that.value)),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_meta_model_description(
    that: MetaModelDescription,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("summary", str(that.summary)),
            stringify.Property("remarks", list(map(str, that.remarks))),
            stringify.Property(
                "constraints_by_identifier",
                [
                    [identifier, str(body)]
                    for identifier, body in that.constraints_by_identifier.items()
                ],
            ),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_symbol_description(
    that: SymbolDescription,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("summary", str(that.summary)),
            stringify.Property("remarks", list(map(str, that.remarks))),
            stringify.Property(
                "constraints_by_identifier",
                [
                    [identifier, str(body)]
                    for identifier, body in that.constraints_by_identifier.items()
                ],
            ),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_property_description(
    that: PropertyDescription,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("summary", str(that.summary)),
            stringify.Property("remarks", list(map(str, that.remarks))),
            stringify.Property(
                "constraints_by_identifier",
                [
                    [identifier, str(body)]
                    for identifier, body in that.constraints_by_identifier.items()
                ],
            ),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_enumeration_literal_description(
    that: EnumerationLiteralDescription,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("summary", str(that.summary)),
            stringify.Property("remarks", list(map(str, that.remarks))),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_signature_description(
    that: SignatureDescription,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("summary", str(that.summary)),
            stringify.Property("remarks", list(map(str, that.remarks))),
            stringify.Property(
                "arguments_by_name",
                [[name, str(body)] for name, body in that.arguments_by_name.items()],
            ),
            stringify.Property(
                "returns", None if that.returns is None else str(that.returns)
            ),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_property(
    that: Property,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("type_annotation", _stringify(that.type_annotation)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property(
                "specified_for",
                f"Reference to {that.specified_for.__class__.__name__} "
                f"{that.specified_for.name}",
            ),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_default_constant(
    that: DefaultConstant,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("value", that.value),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_default_enumeration_literal(
    that: DefaultEnumerationLiteral,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property(
                "enumeration",
                f"Reference to {that.enumeration.__class__.__name__} "
                f"{that.enumeration.name}",
            ),
            stringify.Property(
                "literal",
                f"Reference to {that.literal.__class__.__name__} "
                f"{that.literal.name}",
            ),
        ],
    )

    return result


def _stringify_argument(
    that: Argument,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("type_annotation", _stringify(that.type_annotation)),
            stringify.Property("default", _stringify(that.default)),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_serialization(
    that: Serialization,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("with_model_type", that.with_model_type),
        ],
    )

    return result


def _stringify_reference_in_the_book(
    that: ReferenceInTheBook,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("section", list(that.section)),
            stringify.Property("index", that.index),
            stringify.Property("fragment", that.fragment),
        ],
    )

    return result


def _stringify_invariant(
    that: Invariant,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("description", that.description),
            stringify.Property("body", parse_tree.dump(that.body)),
            stringify.Property(
                "specified_for",
                f"Reference to {that.specified_for.__class__.__name__} "
                f"{that.specified_for.name}",
            ),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_contract(
    that: Contract,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("args", that.args),
            stringify.Property("description", that.description),
            stringify.Property("body", parse_tree.dump(that.body)),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_snapshot(
    that: Snapshot,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("args", that.args),
            stringify.Property("body", parse_tree.dump(that.body)),
            stringify.Property("name", that.name),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_contracts(that: Contracts) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property(
                "preconditions", list(map(_stringify, that.preconditions))
            ),
            stringify.Property("snapshots", list(map(_stringify, that.snapshots))),
            stringify.Property(
                "postconditions",
                list(map(_stringify, that.postconditions)),
            ),
        ],
    )

    return result


def _stringify_a_signature_like(that: SignatureLike) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("arguments", list(map(_stringify, that.arguments))),
            stringify.Property("returns", _stringify(that.returns)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("contracts", _stringify(that.contracts)),
            stringify.PropertyEllipsis("parsed", that.parsed),
            stringify.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_method(that: Method) -> stringify.Entity:
    return _stringify_a_signature_like(that)


def _stringify_implementation_specific_method(
    that: ImplementationSpecificMethod,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("arguments", list(map(_stringify, that.arguments))),
            stringify.Property("returns", _stringify(that.returns)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("contracts", _stringify(that.contracts)),
            stringify.PropertyEllipsis("parsed", that.parsed),
            stringify.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_understood_method(
    that: UnderstoodMethod,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("arguments", list(map(_stringify, that.arguments))),
            stringify.Property("returns", _stringify(that.returns)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("contracts", _stringify(that.contracts)),
            stringify.PropertyEllipsis("parsed", that.parsed),
            stringify.Property("body", [parse_tree.dump(stmt) for stmt in that.body]),
            stringify.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_constructor(
    that: Constructor,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("arguments", list(map(_stringify, that.arguments))),
            stringify.Property("returns", _stringify(that.returns)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("contracts", _stringify(that.contracts)),
            stringify.PropertyEllipsis("parsed", that.parsed),
            stringify.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
            stringify.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify.Property(
                "statements", list(map(construction.dump, that.statements))
            ),
        ],
    )

    return result


def _stringify_enumeration_literal(
    that: EnumerationLiteral,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("value", that.value),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_enumeration(
    that: Enumeration,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("literals", list(map(_stringify, that.literals))),
            stringify.Property(
                "is_superset_of",
                [
                    f"Reference to {parent_enum.__class__.__name__} "
                    f"{parent_enum.name}"
                    for parent_enum in that.is_superset_of
                ],
            ),
            stringify.Property(
                "reference_in_the_book", _stringify(that.reference_in_the_book)
            ),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("literals_by_name", that.literals_by_name),
            stringify.PropertyEllipsis("literal_id_set", that.literal_id_set),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_constrained_primitive(
    that: ConstrainedPrimitive,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property(
                "inheritances",
                [
                    f"Reference to {inheritance.__class__.__name__} {inheritance.name}"
                    for inheritance in that.inheritances
                ],
            ),
            stringify.PropertyEllipsis("inheritance_id_set", that.inheritance_id_set),
            stringify.PropertyEllipsis("descendant_id_set", that.descendant_id_set),
            stringify.Property("constrainee", that.constrainee.name),
            stringify.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify.Property("invariants", list(map(_stringify, that.invariants))),
            stringify.PropertyEllipsis("invariant_id_set", that.invariant_id_set),
            stringify.Property(
                "reference_in_the_book", _stringify(that.reference_in_the_book)
            ),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("parsed", that.parsed),
        ],
    )

    return result


def _stringify_a_class(that: Class) -> stringify.Entity:
    # NOTE (mristin, 2021-12-26):
    # Concrete and abstract class share all the attributes, so we provide a single
    # stringification function.

    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property(
                "inheritances",
                [
                    f"Reference to {inheritance.__class__.__name__} {inheritance.name}"
                    for inheritance in that.inheritances
                ],
            ),
            stringify.PropertyEllipsis("inheritance_id_set", that.inheritance_id_set),
            stringify.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify.Property("interface", _stringify(that.interface)),
            stringify.PropertyEllipsis("descendant_id_set", that.descendant_id_set),
            stringify.Property(
                "concrete_descendants",
                [
                    f"Reference to {descendant.__class__.__name__} {descendant.name}"
                    for descendant in that.concrete_descendants
                ],
            ),
            stringify.Property("properties", list(map(_stringify, that.properties))),
            stringify.Property("methods", list(map(_stringify, that.methods))),
            stringify.Property("constructor", _stringify(that.constructor)),
            stringify.Property("invariants", list(map(_stringify, that.invariants))),
            stringify.Property("serialization", _stringify(that.serialization)),
            stringify.Property(
                "reference_in_the_book", _stringify(that.reference_in_the_book)
            ),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("parsed", that.parsed),
            stringify.PropertyEllipsis("properties_by_name", that.properties_by_name),
            stringify.PropertyEllipsis("property_id_set", that.property_id_set),
            stringify.PropertyEllipsis("methods_by_name", that.methods_by_name),
            stringify.PropertyEllipsis("invariant_id_set", that.invariant_id_set),
        ],
    )

    return result


def _stringify_concrete_class(that: ConcreteClass) -> stringify.Entity:
    return _stringify_a_class(that)


def _stringify_abstract_class(that: AbstractClass) -> stringify.Entity:
    return _stringify_a_class(that)


def _stringify_implementation_specific_verification(
    that: ImplementationSpecificVerification,
) -> stringify.Entity:
    return _stringify_a_signature_like(that)


def _stringify_pattern_verification(
    that: PatternVerification,
) -> stringify.Entity:
    signature_like = _stringify_a_signature_like(that)

    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=list(signature_like.properties)
        + [stringify.Property("pattern", that.pattern)],
    )

    return result


def _stringify_signature(
    that: Signature,
) -> stringify.Entity:
    return _stringify_a_signature_like(that)


def _stringify_interface(
    that: Interface,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property(
                "base",
                f"Reference to {that.base.__class__.__name__} " f"{that.base.name}",
            ),
            stringify.Property("name", that.name),
            stringify.Property(
                "inheritances",
                [
                    f"Reference to {inheritance.__class__.__name__} {inheritance.name}"
                    for inheritance in that.inheritances
                ],
            ),
            stringify.Property(
                "implementers",
                [
                    f"Reference to {implementer.__class__.__name__} {implementer.name}"
                    for implementer in that.implementers
                ],
            ),
            stringify.Property("properties", list(map(_stringify, that.properties))),
            stringify.Property("signatures", list(map(_stringify, that.signatures))),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("parsed", that.parsed),
            stringify.PropertyEllipsis("properties_by_name", that.properties_by_name),
            stringify.PropertyEllipsis("property_id_set", that.property_id_set),
        ],
    )

    return result


def _stringify_meta_model(
    that: MetaModel,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("book_url", that.book_url),
            stringify.Property("book_version", that.book_version),
        ],
    )

    return result


def _stringify_symbol_table(
    that: SymbolTable,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("symbols", list(map(_stringify, that.symbols))),
            stringify.Property(
                "symbols_topologically_sorted",
                [
                    f"Reference to symbol {symbol.name}"
                    for symbol in that.symbols_topologically_sorted
                ],
            ),
            stringify.Property(
                "verification_functions",
                list(map(_stringify, that.verification_functions)),
            ),
            stringify.PropertyEllipsis(
                "verification_functions_by_name", that.verification_functions_by_name
            ),
            stringify.Property("meta_model", _stringify(that.meta_model)),
        ],
    )

    return result


Dumpable = Union[
    AbstractClass,
    Argument,
    ConcreteClass,
    Constructor,
    Contract,
    Contracts,
    Default,
    MetaModelDescription,
    SymbolDescription,
    PropertyDescription,
    EnumerationLiteralDescription,
    SignatureDescription,
    Enumeration,
    EnumerationLiteral,
    ImplementationSpecificMethod,
    ImplementationSpecificVerification,
    Interface,
    Invariant,
    ListTypeAnnotation,
    MetaModel,
    OptionalTypeAnnotation,
    OurTypeAnnotation,
    PatternVerification,
    PrimitiveTypeAnnotation,
    Property,
    ReferenceInTheBook,
    Serialization,
    Signature,
    Snapshot,
    Symbol,
    SymbolTable,
    UnderstoodMethod,
]

stringify.assert_all_public_types_listed_as_dumpables(
    dumpable=Dumpable, types_module=_types
)

_DISPATCH = {
    AbstractClass: _stringify_abstract_class,
    Argument: _stringify_argument,
    ConcreteClass: _stringify_concrete_class,
    ConstrainedPrimitive: _stringify_constrained_primitive,
    Constructor: _stringify_constructor,
    Contract: _stringify_contract,
    Contracts: _stringify_contracts,
    DefaultConstant: _stringify_default_constant,
    DefaultEnumerationLiteral: _stringify_default_enumeration_literal,
    MetaModelDescription: _stringify_meta_model_description,
    SymbolDescription: _stringify_symbol_description,
    PropertyDescription: _stringify_property_description,
    EnumerationLiteralDescription: _stringify_enumeration_literal_description,
    SignatureDescription: _stringify_signature_description,
    Enumeration: _stringify_enumeration,
    EnumerationLiteral: _stringify_enumeration_literal,
    ImplementationSpecificMethod: _stringify_implementation_specific_method,
    ImplementationSpecificVerification: _stringify_implementation_specific_verification,
    Interface: _stringify_interface,
    Invariant: _stringify_invariant,
    ListTypeAnnotation: _stringify_list_type_annotation,
    MetaModel: _stringify_meta_model,
    OptionalTypeAnnotation: _stringify_optional_type_annotation,
    OurTypeAnnotation: _stringify_our_type_annotation,
    PatternVerification: _stringify_pattern_verification,
    PrimitiveTypeAnnotation: _stringify_primitive_type_annotation,
    Property: _stringify_property,
    ReferenceInTheBook: _stringify_reference_in_the_book,
    Serialization: _stringify_serialization,
    Signature: _stringify_signature,
    Snapshot: _stringify_snapshot,
    SymbolTable: _stringify_symbol_table,
    UnderstoodMethod: _stringify_understood_method,
}

stringify.assert_dispatch_exhaustive(dispatch=_DISPATCH, dumpable=Dumpable)


def _stringify(that: Optional[Dumpable]) -> Optional[stringify.Entity]:
    """Dispatch to the correct ``_stringify_*`` method."""
    if that is None:
        return None

    stringify_func = _DISPATCH.get(that.__class__, None)
    if stringify_func is None:
        raise AssertionError(
            f"No stringify function could be found for the class {that.__class__}"
        )

    stringified = stringify_func(that)  # type: ignore
    assert isinstance(stringified, stringify.Entity)
    stringify.assert_compares_against_dict(stringified, that)

    return stringified


def dump(that: Optional[Dumpable]) -> str:
    """Produce a string representation of the ``dumpable`` for testing or debugging."""
    if that is None:
        return repr(None)

    stringified = _stringify(that)
    return stringify.dump(stringified)



# pylint: disable=missing-docstring

import textwrap
import unittest
from typing import Sequence, List

import tests.description

from aas_core_codegen.rdf_shacl import _description as rdf_shacl_description


def tokens_are_equal(
    that: Sequence[rdf_shacl_description.Token],
    other: Sequence[rdf_shacl_description.Token],
) -> bool:
    if len(that) != len(other):
        return False

    for me, you in zip(that, other):
        if type(me) != type(you):  # pylint: disable=unidiomatic-typecheck
            return False

        if isinstance(me, rdf_shacl_description.TokenText):
            assert isinstance(you, rdf_shacl_description.TokenText)
            if me.content != you.content:
                return False

    return True


class Test_renderer(unittest.TestCase):
    def test_bullet_list(self) -> None:
        text = textwrap.dedent(
            """\
            Some bullets:

            * Test
              me
            * Well
            """
        )
        doc = tests.description.parse_restructured_text(text)

        renderer = rdf_shacl_description.Renderer()
        got, error = renderer.transform(element=doc)
        assert error is None, f"{error}="
        assert got is not None

        expected = [
            rdf_shacl_description.TokenText("Some bullets:"),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("* "),
            rdf_shacl_description.TokenText("Test me"),
            rdf_shacl_description.TokenLineBreak(),
            rdf_shacl_description.TokenText("* "),
            rdf_shacl_description.TokenText("Well"),
        ]
        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )

    def test_two_consecutive_notes(self) -> None:
        text = textwrap.dedent(
            """\
            Some text.

            .. note::

                A single paragraph in the note.

            .. note::

                Another note.

                Yet another paragraph.

            Paragraph at root.
            """
        )

        doc = tests.description.parse_restructured_text(text)
        renderer = rdf_shacl_description.Renderer()
        got, error = renderer.transform(element=doc)

        assert error is None, f"{error}="
        assert got is not None

        expected = [
            rdf_shacl_description.TokenText("Some text."),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("NOTE:"),
            rdf_shacl_description.TokenLineBreak(),
            rdf_shacl_description.TokenText("A single paragraph in the note."),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("NOTE:"),
            rdf_shacl_description.TokenLineBreak(),
            rdf_shacl_description.TokenText("Another note."),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("Yet another paragraph."),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("Paragraph at root."),
        ]
        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )


class Test_without_redundant_breaks(unittest.TestCase):
    def test_empty(self) -> None:
        got = rdf_shacl_description.without_redundant_breaks([])
        expected = []  # type: List[rdf_shacl_description.Token]
        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )

    def test_text(self) -> None:
        got = rdf_shacl_description.without_redundant_breaks(
            [rdf_shacl_description.TokenText("something")]
        )

        expected = [rdf_shacl_description.TokenText("something")]

        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )

    def test_consecutive_line_breaks(self) -> None:
        got = rdf_shacl_description.without_redundant_breaks(
            [
                rdf_shacl_description.TokenText("something"),
                rdf_shacl_description.TokenLineBreak(),
                rdf_shacl_description.TokenLineBreak(),
                rdf_shacl_description.TokenText("else"),
            ]
        )

        expected = [
            rdf_shacl_description.TokenText("something"),
            rdf_shacl_description.TokenLineBreak(),
            rdf_shacl_description.TokenText("else"),
        ]

        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )

    def test_consecutive_paragraph_breaks(self) -> None:
        got = rdf_shacl_description.without_redundant_breaks(
            [
                rdf_shacl_description.TokenText("something"),
                rdf_shacl_description.TokenParagraphBreak(),
                rdf_shacl_description.TokenParagraphBreak(),
                rdf_shacl_description.TokenText("else"),
            ]
        )

        expected = [
            rdf_shacl_description.TokenText("something"),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("else"),
        ]

        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )

    def test_trailing_breaks(self) -> None:
        got = rdf_shacl_description.without_redundant_breaks(
            [
                rdf_shacl_description.TokenText("something"),
                rdf_shacl_description.TokenParagraphBreak(),
                rdf_shacl_description.TokenLineBreak(),
            ]
        )

        expected = [rdf_shacl_description.TokenText("something")]

        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )

    def test_all_breaks(self) -> None:
        got = rdf_shacl_description.without_redundant_breaks(
            [
                rdf_shacl_description.TokenParagraphBreak(),
                rdf_shacl_description.TokenLineBreak(),
            ]
        )

        expected = []  # type: List[rdf_shacl_description.Token]

        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )

    def test_regression_on_skipped_tokens(self) -> None:
        # This is an actual bug that caused unexpected cut-off of the tokens.
        tokens = [
            rdf_shacl_description.TokenText("An element that is referable by its "),
            rdf_shacl_description.TokenText("'idShort'"),
            rdf_shacl_description.TokenText("."),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("Something."),
            rdf_shacl_description.TokenParagraphBreak(),
        ]  # type: List[rdf_shacl_description.TokenUnion]

        got = rdf_shacl_description.without_redundant_breaks(tokens)

        expected = [
            rdf_shacl_description.TokenText("An element that is referable by its "),
            rdf_shacl_description.TokenText("'idShort'"),
            rdf_shacl_description.TokenText("."),
            rdf_shacl_description.TokenParagraphBreak(),
            rdf_shacl_description.TokenText("Something."),
        ]

        self.assertTrue(
            tokens_are_equal(expected, got), f"Expected {expected!r}, got {got!r}"
        )


if __name__ == "__main__":
    unittest.main()



"""Generate different implementations and schemas based on an AAS meta-model."""

import argparse
import enum
import pathlib
import sys
from typing import TextIO

import aas_core_codegen
from aas_core_codegen import parse, run, specific_implementations, intermediate
from aas_core_codegen.common import LinenoColumner, assert_never
import aas_core_codegen.csharp.main as csharp_main
import aas_core_codegen.jsonschema.main as jsonschema_main
import aas_core_codegen.rdf_shacl.main as rdf_shacl_main

assert aas_core_codegen.__doc__ == __doc__


class Target(enum.Enum):
    """List available target implementations."""

    CSHARP = "csharp"
    JSONSCHEMA = "jsonschema"
    RDF_SHACL = "rdf_shacl"


class Parameters:
    """Represent the program parameters."""

    def __init__(
        self,
        model_path: pathlib.Path,
        target: Target,
        snippets_dir: pathlib.Path,
        output_dir: pathlib.Path,
    ) -> None:
        """Initialize with the given values."""
        self.model_path = model_path
        self.target = target
        self.snippets_dir = snippets_dir
        self.output_dir = output_dir


# noinspection SpellCheckingInspection
def execute(params: Parameters, stdout: TextIO, stderr: TextIO) -> int:
    """Run the program."""
    # region Basic checks
    # BEFORE-RELEASE (mristin, 2021-12-13): test this failure case
    if not params.model_path.exists():
        stderr.write(f"The --model_path does not exist: {params.model_path}\n")
        return 1

    # BEFORE-RELEASE (mristin, 2021-12-13): test this failure case
    if not params.model_path.is_file():
        stderr.write(
            f"The --model_path does not point to a file: {params.model_path}\n"
        )
        return 1

    # BEFORE-RELEASE (mristin, 2021-12-13): test this failure case
    if not params.snippets_dir.exists():
        stderr.write(f"The --snippets_dir does not exist: {params.snippets_dir}\n")
        return 1

    # BEFORE-RELEASE (mristin, 2021-12-13): test this failure case
    if not params.snippets_dir.is_dir():
        stderr.write(
            f"The --snippets_dir does not point to a directory: "
            f"{params.snippets_dir}\n"
        )
        return 1

    # BEFORE-RELEASE (mristin, 2021-12-13): test the happy path
    if not params.output_dir.exists():
        params.output_dir.mkdir(parents=True, exist_ok=True)
    else:
        # BEFORE-RELEASE (mristin, 2021-12-13): test this failure case
        if not params.output_dir.is_dir():
            stderr.write(
                f"The --output_dir does not point to a directory: "
                f"{params.output_dir}\n"
            )
            return 1

    # endregion

    # region Parse

    spec_impls, spec_impls_errors = specific_implementations.read_from_directory(
        snippets_dir=params.snippets_dir
    )

    if spec_impls_errors:
        run.write_error_report(
            message="Failed to resolve the implementation-specific "
            "JSON schema snippets",
            errors=spec_impls_errors,
            stderr=stderr,
        )
        return 1

    assert spec_impls is not None

    text = params.model_path.read_text(encoding="utf-8")

    # BEFORE-RELEASE (mristin, 2021-12-13):
    #  test all the following individual failure cases
    atok, parse_exception = parse.source_to_atok(source=text)
    if parse_exception:
        if isinstance(parse_exception, SyntaxError):
            stderr.write(
                f"Failed to parse the meta-model {params.model_path}: "
                f"invalid syntax at line {parse_exception.lineno}\n"
            )
        else:
            stderr.write(
                f"Failed to parse the meta-model {params.model_path}: "
                f"{parse_exception}\n"
            )

        return 1

    import_errors = parse.check_expected_imports(atok=atok)
    if import_errors:
        run.write_error_report(
            message="One or more unexpected imports in the meta-model",
            errors=import_errors,
            stderr=stderr,
        )

        return 1

    lineno_columner = LinenoColumner(atok=atok)

    parsed_symbol_table, error = parse.atok_to_symbol_table(atok=atok)
    if error is not None:
        run.write_error_report(
            message=f"Failed to construct the symbol table from {params.model_path}",
            errors=[lineno_columner.error_message(error)],
            stderr=stderr,
        )

        return 1

    assert parsed_symbol_table is not None

    ir_symbol_table, error = intermediate.translate(
        parsed_symbol_table=parsed_symbol_table,
        atok=atok,
    )
    if error is not None:
        run.write_error_report(
            message=f"Failed to translate the parsed symbol table "
            f"to intermediate symbol table "
            f"based on {params.model_path}",
            errors=[lineno_columner.error_message(error)],
            stderr=stderr,
        )

        return 1

    assert ir_symbol_table is not None

    # endregion

    # region Dispatch

    run_context = run.Context(
        model_path=params.model_path,
        symbol_table=ir_symbol_table,
        spec_impls=spec_impls,
        lineno_columner=lineno_columner,
        output_dir=params.output_dir,
    )

    if params.target is Target.CSHARP:
        return csharp_main.execute(context=run_context, stdout=stdout, stderr=stderr)

    elif params.target is Target.JSONSCHEMA:
        return jsonschema_main.execute(
            context=run_context, stdout=stdout, stderr=stderr
        )

    elif params.target is Target.RDF_SHACL:
        return rdf_shacl_main.execute(context=run_context, stdout=stdout, stderr=stderr)

    else:
        assert_never(params.target)

    # endregion

    stdout.write(f"Code generated to: {params.output_dir}\n")
    return 0


def main(prog: str) -> int:
    """
    Execute the main routine.

    :param prog: name of the program to be displayed in the help
    :return: exit code
    """
    parser = argparse.ArgumentParser(prog=prog, description=__doc__)
    parser.add_argument("--model_path", help="path to the meta-model", required=True)
    parser.add_argument(
        "--snippets_dir",
        help="path to the directory containing implementation-specific code snippets",
        required=True,
    )
    parser.add_argument(
        "--output_dir", help="path to the generated code", required=True
    )
    parser.add_argument(
        "--target",
        help="target language or schema",
        required=True,
        choices=[literal.value for literal in Target],
    )
    parser.add_argument(
        "--version", help="show the current version and exit", action="store_true"
    )

    # NOTE (mristin, 2022-01-14):
    # The module ``argparse`` is not flexible enough to understand special options such
    # as ``--version`` so we manually hard-wire.
    if "--version" in sys.argv and "--help" not in sys.argv:
        print(aas_core_codegen.__version__)
        return 1

    args = parser.parse_args()

    target_to_str = {literal.value: literal for literal in Target}

    params = Parameters(
        model_path=pathlib.Path(args.model_path),
        target=target_to_str[args.target],
        snippets_dir=pathlib.Path(args.snippets_dir),
        output_dir=pathlib.Path(args.output_dir),
    )

    return execute(params=params, stdout=sys.stdout, stderr=sys.stderr)


def entry_point() -> int:
    """Provide an entry point for a console script."""
    return main(prog="aas-core-codegen")


if __name__ == "__main__":
    sys.exit(main(prog="aas-core-codegen"))



"""Merge constrained primitives as property constraints."""
import collections
from typing import Tuple, Optional, List, Mapping, MutableMapping, Sequence

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Error
from aas_core_codegen.infer_for_schema import (
    _len as infer_for_schema_len,
    _pattern as infer_for_schema_pattern,
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _infer_len_constraints_by_constrained_primitive(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[
    Optional[
        MutableMapping[
            intermediate.ConstrainedPrimitive, infer_for_schema_len.LenConstraint
        ]
    ],
    Optional[List[Error]],
]:
    """Infer the constraints on ``len(.)`` of the constrained primitives."""

    # NOTE (mristin, 2022-02-11):
    # We do this inference in two passes. In the first pass, we only infer
    # the constraints defined for the constrained primitive and ignore the ancestors.
    # In the second pass, we stack the constraints of the ancestors as well.

    errors = []  # type: List[Error]

    first_pass: MutableMapping[
        intermediate.ConstrainedPrimitive, infer_for_schema_len.LenConstraint
    ] = dict()

    for symbol in symbol_table.symbols:
        if isinstance(symbol, intermediate.ConstrainedPrimitive):
            (
                len_constraint,
                len_constraint_errors,
            ) = infer_for_schema_len.infer_len_constraint_of_self(
                constrained_primitive=symbol
            )

            if len_constraint_errors is not None:
                errors.extend(len_constraint_errors)
            else:
                assert len_constraint is not None

                first_pass[symbol] = len_constraint

    if len(errors) > 0:
        return None, errors

    second_pass: MutableMapping[
        intermediate.ConstrainedPrimitive, infer_for_schema_len.LenConstraint
    ] = dict()

    for symbol in symbol_table.symbols_topologically_sorted:
        if isinstance(symbol, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-02-11):
            # We make the copy in order to avoid bugs when we start processing
            # the inheritances.
            len_constraint = first_pass[symbol].copy()

            for inheritance in symbol.inheritances:
                inherited_len_constraint = second_pass.get(inheritance, None)
                assert (
                    inherited_len_constraint is not None
                ), "Expected topological order"

                if inherited_len_constraint.min_value is not None:
                    len_constraint.min_value = (
                        max(
                            len_constraint.min_value, inherited_len_constraint.min_value
                        )
                        if len_constraint.min_value is not None
                        else inherited_len_constraint.min_value
                    )

                if inherited_len_constraint.max_value is not None:
                    len_constraint.max_value = (
                        min(
                            len_constraint.max_value, inherited_len_constraint.max_value
                        )
                        if len_constraint.max_value is not None
                        else inherited_len_constraint.max_value
                    )

            second_pass[symbol] = len_constraint

    assert len(errors) == 0
    return second_pass, None


def _infer_pattern_constraints_by_constrained_primitive(
    symbol_table: intermediate.SymbolTable,
    pattern_verifications_by_name: infer_for_schema_pattern.PatternVerificationsByName,
) -> MutableMapping[
    intermediate.ConstrainedPrimitive, List[infer_for_schema_pattern.PatternConstraint]
]:
    """Infer the pattern constraints of the constrained strings."""

    # NOTE (mristin, 2022-02-11):
    # We do this inference in two passes. In the first pass, we only infer
    # the constraints defined for the constrained primitive and ignore the ancestors.
    # In the second pass, we stack the constraints of the ancestors as well.

    first_pass: MutableMapping[
        intermediate.ConstrainedPrimitive,
        List[infer_for_schema_pattern.PatternConstraint],
    ] = collections.OrderedDict()

    for symbol in symbol_table.symbols:
        if (
            isinstance(symbol, intermediate.ConstrainedPrimitive)
            and symbol.constrainee is intermediate.PrimitiveType.STR
        ):
            pattern_constraints = infer_for_schema_pattern.infer_patterns_on_self(
                constrained_primitive=symbol,
                pattern_verifications_by_name=pattern_verifications_by_name,
            )

            first_pass[symbol] = pattern_constraints

    second_pass: MutableMapping[
        intermediate.ConstrainedPrimitive,
        List[infer_for_schema_pattern.PatternConstraint],
    ] = dict()

    for symbol in first_pass:
        # NOTE (mristin, 2022-02-11):
        # We make the copy in order to avoid bugs when we start processing
        # the inheritances.
        pattern_constraints = first_pass[symbol][:]

        for inheritance in symbol.inheritances:
            assert inheritance in first_pass, (
                f"We are processing the constrained primitive {symbol.name!r}. "
                f"However, its parent, {inheritance.name!r}, has not been processed in "
                f"the first pass. Something probably went wrong in the first pass."
            )

            inherited_pattern_constraints = second_pass.get(inheritance, None)
            assert inherited_pattern_constraints is not None, (
                f"Expected topological order. However, the symbol {symbol.name!r} "
                f"is being processed before one of its parents, {inheritance.name!r}."
            )

            pattern_constraints = inherited_pattern_constraints + pattern_constraints

        second_pass[symbol] = pattern_constraints

    return second_pass


class ConstraintsByProperty:
    """
    Represent all the inferred property constraints of a symbol.

    The constraints coming from the constrained primitives are in-lined and hence also
    included in this representation.
    """

    def __init__(
        self,
        len_constraints_by_property: Mapping[
            intermediate.Property, infer_for_schema_len.LenConstraint
        ],
        patterns_by_property: Mapping[
            intermediate.Property, Sequence[infer_for_schema_pattern.PatternConstraint]
        ],
    ) -> None:
        """Initialize with the given values."""
        self.len_constraints_by_property = len_constraints_by_property
        self.patterns_by_property = patterns_by_property


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def infer_constraints_by_class(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[
    Optional[MutableMapping[intermediate.ClassUnion, ConstraintsByProperty]],
    Optional[List[Error]],
]:
    """Infer the constraints from the invariants and constrained primitives."""
    errors = []  # type: List[Error]

    pattern_verifications_by_name = (
        infer_for_schema_pattern.map_pattern_verifications_by_name(
            verifications=symbol_table.verification_functions
        )
    )

    (
        len_constraints_by_constrained_primitive,
        some_errors,
    ) = _infer_len_constraints_by_constrained_primitive(symbol_table=symbol_table)
    if some_errors is not None:
        errors.extend(some_errors)

    if len(errors) > 0:
        return None, errors

    assert len_constraints_by_constrained_primitive is not None

    patterns_by_constrained_primitive = (
        _infer_pattern_constraints_by_constrained_primitive(
            symbol_table=symbol_table,
            pattern_verifications_by_name=pattern_verifications_by_name,
        )
    )

    result: MutableMapping[intermediate.ClassUnion, ConstraintsByProperty] = dict()

    for symbol in symbol_table.symbols:
        if not isinstance(
            symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            continue

        # region Infer constraints on ``len(.)``

        len_constraints_by_property: MutableMapping[
            intermediate.Property, infer_for_schema_len.LenConstraint
        ] = dict()

        (
            len_constraints_from_invariants,
            len_constraints_errors,
        ) = infer_for_schema_len.len_constraints_from_invariants(cls=symbol)

        if len_constraints_errors is not None:
            errors.extend(len_constraints_errors)
            continue

        assert len_constraints_from_invariants is not None

        patterns_by_property: MutableMapping[
            intermediate.Property, List[infer_for_schema_pattern.PatternConstraint]
        ] = dict()

        patterns_from_invariants_by_property = (
            infer_for_schema_pattern.patterns_from_invariants(
                cls=symbol, pattern_verifications_by_name=pattern_verifications_by_name
            )
        )

        for prop in symbol.properties:
            # NOTE (mristin, 2022-03-3):
            # We need to go beneath ``Optional`` as the constraints are applied even
            # if a property is optional. In cases where cardinality is affected by
            # ``Optional``, the client code needs to cover them separately.
            type_anno = intermediate.beneath_optional(prop.type_annotation)

            len_constraint_from_type: Optional[
                infer_for_schema_len.LenConstraint
            ] = None

            len_constraint_from_invariants = len_constraints_from_invariants.get(
                prop, None
            )

            if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
                type_anno.symbol, intermediate.ConstrainedPrimitive
            ):
                len_constraint_from_type = len_constraints_by_constrained_primitive.get(
                    type_anno.symbol, None
                )

            # Merge the constraint from the type and from the invariants

            if (
                len_constraint_from_type is None
                and len_constraint_from_invariants is None
            ):
                continue

            elif (
                len_constraint_from_type is not None
                and len_constraint_from_invariants is None
            ):
                len_constraints_by_property[prop] = len_constraint_from_type

            elif (
                len_constraint_from_type is None
                and len_constraint_from_invariants is not None
            ):
                len_constraints_by_property[prop] = len_constraint_from_invariants

            elif (
                len_constraint_from_type is not None
                and len_constraint_from_invariants is not None
            ):
                # NOTE (mristin, 2022-03-2):
                # We have to make the bounds *stricter* since both
                # the type constraints and the invariant(s) need to be satisfied.

                min_value = infer_for_schema_len.max_with_none(
                    len_constraint_from_type.min_value,
                    len_constraint_from_invariants.min_value,
                )

                max_value = infer_for_schema_len.min_with_none(
                    len_constraint_from_type.max_value,
                    len_constraint_from_invariants.max_value,
                )

                if (
                    min_value is not None
                    and max_value is not None
                    and min_value > max_value
                ):
                    errors.append(
                        Error(
                            symbol.parsed.node,
                            f"The inferred minimum and maximum value on len(.) "
                            f"is contradictory: "
                            f"minimum = {min_value}, maximum = {max_value}; "
                            f"please check the invariants and "
                            f"any involved constrained primitives",
                        )
                    )
                    continue

                len_constraints_by_property[prop] = infer_for_schema_len.LenConstraint(
                    min_value=min_value, max_value=max_value
                )

            else:
                raise AssertionError(
                    f"Unhandled case: "
                    f"{len_constraint_from_type=}, {len_constraint_from_invariants}"
                )

            # endregion

            # region Infer constraints on string patterns

            patterns_from_type: List[infer_for_schema_pattern.PatternConstraint] = []
            patterns_from_invariants = patterns_from_invariants_by_property.get(
                prop, []
            )

            if isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
                type_anno.symbol, intermediate.ConstrainedPrimitive
            ):
                patterns_from_type = patterns_by_constrained_primitive.get(
                    type_anno.symbol, []
                )

            merged = patterns_from_type + patterns_from_invariants

            if len(merged) > 0:
                patterns_by_property[prop] = merged

        # endregion

        if len(errors) > 0:
            return None, errors

        result[symbol] = ConstraintsByProperty(
            len_constraints_by_property=len_constraints_by_property,
            patterns_by_property=patterns_by_property,
        )

    return result, None
